(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_task_server.ml 239 2010-06-23 16:49:03Z gerd $ *)

type sh_state =
    { pid_of_req_id : (int, int) Hashtbl.t }

module State_shvar =
  Netplex_sharedvar.Make_var_type(struct type t = sh_state end)

let state_name = "mapred_task_server.state"
let mutex_name = state_name ^ ".mutex"

let m = ref None

let change_state op =
  let mutex = Netplex_mutex.access mutex_name in
  Netplex_mutex.lock mutex;
  try
    ( match op with
	| `Reg_req(req_id,pid) ->
	    let st = State_shvar.get state_name in
	    Hashtbl.replace st.pid_of_req_id req_id pid;
	    State_shvar.set state_name st
	| `Del_req req_id ->
	    let st = State_shvar.get state_name in
	    Hashtbl.remove st.pid_of_req_id req_id;
	    State_shvar.set state_name st
	| `Init ->
	    let st = 
	      { pid_of_req_id = Hashtbl.create 5 } in
	    let ok =
	      Netplex_sharedvar.create_var ~enc:true state_name in
	    if ok then
	      State_shvar.set state_name st
    );
    Netplex_mutex.unlock mutex;
  with
    | error ->
	Netplex_mutex.unlock mutex;
	raise error

(* needed config:
   - mapred_env:
     * cluster name, namenodes
     * config file
     * arguments (string array) --> exec!
 *)

let proc_exec (mj,conf,cf) (req_id, args, enc_task) =
  try
    change_state (`Reg_req(req_id, Unix.getpid()));
    let shm_mng = new Plasma_shm.netplex_shm_manager() in
    let me =
      ( object
	  method open_cluster esys =
	    let c =
	      Plasma_client.open_cluster 
		conf#nn_clustername
		(List.map Plasma_util.parse_host_port conf#nn_nodes)
		esys in
	    Plasma_client.configure_shm_manager c shm_mng;
	    c
	      
	  method config_file = cf
	  method arguments = args
	end
      ) in
    let task = Mapred_tasks.decode_task enc_task in
    let r = Mapred_task_exec.execute me mj task in
    change_state (`Del_req req_id);
    Mapred_tasks.encode_task_result r
  with
    | error ->
	Netlog.logf `Err "Exception in exec (pid %d): %s"
	  (Unix.getpid()) (Netexn.to_string error);
	Mapred_tasks.encode_task_result
	  (`Error (Netexn.to_string error))


let proc_kill req_id =
  let st = State_shvar.get state_name in
  try
    let pid = Hashtbl.find st.pid_of_req_id req_id in
    ( try
	Unix.kill pid Sys.sigterm
      with
	| error ->
	    Netlog.logf `Err "Cannot kill pid %d: %s"
	      pid
	      (Netexn.to_string error);
    );
    change_state (`Del_req req_id)
  with
    | Not_found ->
	()


let proc_kill_all () =
  let st = State_shvar.get state_name in
  Hashtbl.iter
    (fun req_id pid ->
       ( try
	   Unix.kill pid Sys.sigterm
	 with
	   | error ->
	       Netlog.logf `Err "Cannot kill pid %d: %s"
		 pid
		 (Netexn.to_string error);
       );
       change_state (`Del_req req_id)
    )
    st.pid_of_req_id


let proc_configure (low,high) =
  Plasma_shm.Netplex_shm_manager.init ~low ~high


let setup srv (mj,conf,cf) =
  Mapred_rpc_srv.Mapred_task.V1.bind
    ~proc_null:(fun () -> ())
    ~proc_exec:(proc_exec (mj,conf,cf))
    ~proc_kill
    ~proc_kill_all
    ~proc_configure
    srv;
  Rpc_server.set_exception_handler
    srv
    (fun e ->
       prerr_endline(Netexn.to_string e)
    )


let factory mj job_cf =
  Rpc_netplex.rpc_factory
   ~configure:(fun _ _ -> 
		 (mj, Mapred_config.extract_config job_cf, job_cf))
    ~hooks:(fun _ ->
              ( object
                  inherit Netplex_kit.empty_processor_hooks()
                    
                  method post_add_hook socksrv ctrl =
		    ctrl # add_plugin Netplex_mutex.plugin;
		    ctrl # add_plugin Netplex_sharedvar.plugin		    

                  method post_start_hook cont =
		    change_state `Init;

		    Rpc_server.Debug.enable := true;
		    Rpc_server.Debug.enable_ctrace := true;
		    Rpc_server.Debug.enable_ptrace := true;
		end
	      )
	   )
    ~setup
    ~name:"mapred_task_server"
    ()
