(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_blockmap.ml 239 2010-06-23 16:49:03Z gerd $ *)

(* TODO: return blocks in ascending order *)


open Printf

exception Inactive

class type blockmap_t =
object
  method active : bool
  method id : int
  method identity : string
  method size : int64
  method n_used : int64
  method n_trans : int64
  method n_free : int64
  method fill_blockmap : int64 -> string -> unit
  method has_free_blocks : bool
  method reserve : int -> < > -> int64 list
  method free : int64 list -> < > -> unit
  method keep : int64 list -> < > -> unit
  method commit : < > -> unit
  method rollback : < > -> unit
  method get_changes : < > -> (int64 * string) list
  method inactivate : unit -> unit
  method dump : string
end


let rec mklist p0 p1 =
  if p0 <= p1 then
    p0 :: mklist (p0+1) p1
  else
    []


let bmaprowsize = Nn_config.bmaprowsize
let bmaprowsizeL = Int64.of_int bmaprowsize


module IntSet =
  Set.Make(struct
	     type t = int
	     let compare x y =
	       Pervasives.compare (x:int) (y:int)
	   end
	  )

let count_ones bv =
  Bitv.fold_left (fun acc b -> if b then acc+1 else acc) 0 bv


class blockmap ~id ~identity ~size : blockmap_t =
  let () = assert(size > 0L) in
  let size_rows_64 = Int64.succ (Int64.div (Int64.pred size) bmaprowsizeL) in
  let () =
    (* impl. limitation *)
    if size_rows_64 > Int64.of_int Sys.max_array_length then
      failwith "Nn_state.blockmap: datastore too large" in
  let size_rows = Int64.to_int size_rows_64 in
  let size_rem = Int64.to_int (Int64.rem size bmaprowsizeL) in
  let n_last = if size_rem = 0 then bmaprowsize else size_rem in
object(self)
  val mutable active = true
  val mutable committed_map = [| |]
    (** row k contains the bits of the blocks k*1024 to k*1024+1023
        (except for the last row which may be shorter). The bits are
        stored as {!Bitv} bit vector. Bit '1' means the block is used.
        A '0' leaves it open whether it is free or reserved.

        (1024 = bmaprowsize)
     *)
  val mutable reserve_map = [| |]
    (** same layout as [commited_map]. The bits here are only meaningful
        when the corresponding bit in [committed_map] is 0. In that case
        a '1' means that the block is reserved by some owner and '0'
        means there is no such reservation
     *)
   val mutable keep_map = [| |]
    (** same layout as [commited_map]. The meaning depends on the
	value of the corresponding bit in  [committed_map]:
	- If the bit in [committed_map] is 0, a 1 in [keep_map] means
	  that the block is kept, and that
	  the competing transaction did already commit the [free]
	- If the bit in [committed_map] is 1, a 1 in [keep_map] means
	  that the block is kept, and that there was no committed [free]
	  yet

        Only if all the bits in [committed_map], [reserve_map] and [keep_map]
	are 0, the block is really free and can be allocated again.
     *)
   val mutable stats = [| |]
     (** per-row statistics:

	 {[ stats.(k) = (n_used_k, n_reserved_k, n_free_kept_k, n_alloc_k) ]}

	 where
	 - [n_used_k]: the number of committed used blocks in row [k]
	 - [n_reserved_k]: the number of reserved blocks in row [k]
	 - [n_free_kept_k]: the number of committed free blocks that are
           kept by a transaction in row [k]
	 - [n_alloc_k]: the number of allocatable blocks
      *)
  val mutable n_used = 0L
  val mutable n_reserved = 0L
  val mutable n_free_kept = 0L
    (** #blocks that are free and committed but where there is a keep bit *)
  val mutable alloctab = [| |]
    (** maps the number of allocatable blocks to the set of rows *)
  val mutable alloctab_out_of_date = false
    (** false: have to recompute alloctab, n_used, n_reserved *)
  val mutable owner = Hashtbl.create 7
    (** maps owner objects to changes: see [get_owner_tuple] *)
  val mutable rowowner = Hashtbl.create 7
    (** maps rows to owner objects *)
    

  initializer (
    committed_map <-
      Array.init
	size_rows
	(fun k ->
	   let len = if k=size_rows-1 then n_last else bmaprowsize in
	   Bitv.create len false
	);
    reserve_map <- Array.map Bitv.copy committed_map;
    keep_map <- Array.map Bitv.copy committed_map;
    stats <-
      Array.init
	size_rows
	(fun k ->
	   let len = if k=size_rows-1 then n_last else bmaprowsize in
	   (0,0,0,len)
	);
    alloctab_out_of_date <- true
  )

  method dump =
    let b = Buffer.create 500 in
    for row = 0 to size_rows-1 do
      let crow = committed_map.(row) in
      let rrow = reserve_map.(row) in
      let krow = keep_map.(row) in
      let (used,res,freekept,alloc) = stats.(row) in
      bprintf b "[%d]: (used=%d res=%d freekept=%d alloc=%d)\n"
	row used res freekept alloc;
      for k=0 to  Bitv.length crow-1 do
	let block =
	  Int64.add
	    (Int64.mul (Int64.of_int row) bmaprowsizeL) (Int64.of_int k) in
	if Int64.rem block 64L = 0L then (
	  if k > 0 then bprintf b "\n";
	  bprintf b "%6Ld: " block;
	);
	let c =
	  match Bitv.get crow k, Bitv.get rrow k, Bitv.get krow k with
	    | false, false, false -> '_'
	    | false, false, true  -> 'k'
	    | false, true,  false -> 'r'
	    | true,  false, false -> '*'
	    | true,  false, true  -> '#'
	    | _                   -> '?'  (* illegal *) in
	Buffer.add_char b c
      done;
      bprintf b "\n";
    done;
    Buffer.contents b

  method active = active
  method id = id
  method identity = identity

  method size = 
    if not active then raise Inactive;
    size

  method fill_blockmap idx data =
    if not active then raise Inactive;
    if idx < 0L || idx >= size || Int64.rem idx bmaprowsizeL <> 0L then
      invalid_arg "fill_blockmap";
    let row = Int64.to_int(Int64.div idx bmaprowsizeL) in
    if row >= size_rows then
      invalid_arg "fill_blockmap";
    let l_expected =
      if row = size_rows-1 then
	n_last
      else
	bmaprowsize in
    if String.length data <> l_expected then
      failwith "fill_blockmap";
    let bv = Bitv.of_string data in
    let n1 = Bitv.fold_left (fun n b -> if b then n+1 else n) 0 bv in
    committed_map.(row) <- bv;
    reserve_map.(row) <- Bitv.create l_expected false;
    keep_map.(row) <- Bitv.create l_expected false;
    stats.(row) <- (n1, 0, 0, l_expected-n1);
    alloctab_out_of_date <- true;

  method private fix_alloctab() =
    if alloctab_out_of_date then (
      n_used <- 0L;
      n_reserved <- 0L;
      n_free_kept <- 0L;
      alloctab <- Array.init (bmaprowsize+1) (fun _ -> IntSet.empty);
      Array.iteri
	(fun row (u,r,fk,al) ->
	   alloctab.(al) <- IntSet.add row alloctab.(al);
	   n_used <- Int64.add n_used (Int64.of_int u);
	   n_reserved <- Int64.add n_reserved (Int64.of_int r);
	   n_free_kept <- Int64.add n_free_kept (Int64.of_int fk);
	)
	stats;
      alloctab_out_of_date <- false
    )

  method n_used = 
    if not active then raise Inactive;
    self # fix_alloctab();
    n_used

  method n_trans = 
    if not active then raise Inactive;
    self # fix_alloctab();
    Int64.add n_reserved n_free_kept

  method n_free =
    if not active then raise Inactive;
    Int64.sub (Int64.sub size self#n_used) self#n_trans

  method has_free_blocks =
    if not active then raise Inactive;
    self#n_free > 0L

  method private get_owner_tuple row ownobj =
    (** Returns 
       [(orow_res, orow_keep, orow_or, orow_and, n_res, n_keep, n_or, n_and)].

	Here, [orow_res] says which blocks have once been reserved during
        the transaction (i.e. blocks are kept here that were first allocated
        and then freed). [orow_res] has negative logic: 0=block is reserved.

	[orow_keep] says which blocks are kept during the transaction.
	[orow_keep] has positive logic: 1=block is kept.

        [orow_or] says which blocks are finally allocated (i.e. blocks are
        not in here that were first allocated and then freed in the same
        transaction). [orow_or] has positive logic: 1=block is allocated.

        [orow_and] says which blocks are finally freed (note that you cannot
        undo [free] during a transaction except rolling it back).
        [orow_and] has negative logic: 0=block is freed.

        [n_res], [n_keep], [n_or], [n_and]: Number of blocks in the
        respective bitsets
     *)
    let orow_ht =
      try
	Hashtbl.find owner ownobj
      with
	| Not_found ->
	    let ht = Hashtbl.create 7 in
	    Hashtbl.add owner ownobj ht;
	    ht in
    try Hashtbl.find orow_ht row
    with Not_found ->
      let len = (if row = size_rows-1 then n_last else bmaprowsize) in
      let orow_res = Bitv.create len true in
      let orow_keep = Bitv.create len false in
      let orow_or = Bitv.create len false in
      let orow_and = Bitv.create len true in
      let tuple = (orow_res, orow_keep, orow_or, orow_and,
		   ref 0, ref 0, ref 0, ref 0) in
      Hashtbl.add orow_ht row tuple;
      let rol =
	try Hashtbl.find rowowner row 
	with Not_found -> [] in
      Hashtbl.replace rowowner row (ownobj :: rol);
      tuple

  method private orow_keep ownobj row =  (* may raise Not_found *)
    let orow_ht = Hashtbl.find owner ownobj in
    let (_, orow_keep, _, _, _, _, _, _) = Hashtbl.find orow_ht row in
    orow_keep

  method private remove_owner ownobj =
    try
      let orow_ht = Hashtbl.find owner ownobj in
      Hashtbl.iter
	(fun row _ ->
	   let rol = 
	     try Hashtbl.find rowowner row 
	     with Not_found -> [] in
	   let rol' =
	     List.filter (fun o -> o <> ownobj) rol in
	   if rol' <> [] then
	     Hashtbl.replace rowowner row rol'
	   else
	     Hashtbl.remove rowowner row
	)
	orow_ht;
      Hashtbl.remove owner ownobj;
    with
      | Not_found -> ()

  method reserve n ownobj =
    if not active then raise Inactive;
    self # fix_alloctab();
    if n < 0 then
      invalid_arg "reserve";
    if self#n_free < Int64.of_int n then
      failwith "Nn_state.blockmap: not enough free blocks";
    (** First look for rows with >= n blocks *)
    let n = ref n in  (* remaining *)
    let blocks = ref [] in
    let found = ref true in
    while !found do
      found := false;
      let u = min !n bmaprowsize in
      let j = ref u in
      while !j <= bmaprowsize && !n > 0 do
	let l = alloctab.( !j ) in
	if l <> IntSet.empty then (
	  let bl = self # reserve_at !j u ownobj in
	  blocks := bl @ !blocks;
	  n := !n - u;
	  found := true
	)
	else
	  incr j
      done
    done;
    (** Second strategy: look for rows with up to n blocks *)
    if !n > 0 then (
      let j = ref bmaprowsize in
      while !j > 0 && !n > 0 do
	let l = alloctab.( !j ) in
	let u = min !j !n in
	if l <> IntSet.empty then (
	  let bl = self # reserve_at !j u ownobj in
	  blocks := bl @ !blocks;
	  n := !n - u;
	)
	else
	  decr j
      done;
      assert(!n = 0)
    );
    List.rev !blocks


  method private reserve_at diridx n ownobj =
    (** returns blocks in reverse order! *)
    let row =
      try IntSet.min_elt alloctab.( diridx )
      with Not_found -> assert false in
    (** Reserve [n] blocks in [row] for [ownobj] *)
    let crow = committed_map.(row) in
    let rrow = reserve_map.(row) in
    let krow = keep_map.(row) in
    let (orow_res, orow_keep, orow_or, orow_and, 
	 n_res, n_keep, n_or, n_and) =
      self # get_owner_tuple row ownobj in
    let blocks = ref [] in
    let c = ref 0 in
    for k = 0 to Bitv.length crow - 1 do
      if !c < n && not (Bitv.get crow k) && not (Bitv.get rrow k) && 
        not (Bitv.get krow k) 
      then (
	assert(Bitv.get orow_res k);
	assert(not (Bitv.get orow_or k));
	Bitv.set rrow k true;
	Bitv.set orow_res k false;
	Bitv.set orow_or k true;
	let block = 
	  Int64.add
	    (Int64.mul (Int64.of_int row) bmaprowsizeL) 
	    (Int64.of_int k) in
	blocks := block :: !blocks;
	incr n_or;   (* number of 1 bits in orow_or *)
	incr n_res;  (* number of 0 bits in orow_res *)
	incr c
      )
    done;
    let (u,r,fk,al) = stats.(row) in
    let r' = r + n in
    let al' = al - n in
    stats.(row) <- (u,r',fk,al');
    alloctab.( diridx ) <- IntSet.remove row alloctab.( diridx );
    alloctab.( al' ) <- IntSet.add row alloctab.( al' );
    n_reserved <- Int64.add n_reserved (Int64.of_int n);
    !blocks


  method free l ownobj =
    (** well, a dumb loop. FIXME *)
    if not active then raise Inactive;
    self # fix_alloctab();
    List.iter
      (fun block ->
	 let row = Int64.div block bmaprowsizeL in
	 if row >= size_rows_64 then
	   invalid_arg "Nn_state.blockmap.free";
	 self # free_at (Int64.to_int row) [block] ownobj
      )
      l

  method private free_at row l ownobj =
    (** Free the blocks in [l] which are all in [row] *)
    let crow = committed_map.(row) in
    let rrow = reserve_map.(row) in
    let (orow_res, orow_keep, orow_or, orow_and, 
	 n_res, n_keep, n_or, n_and) =
      self # get_owner_tuple row ownobj in
    List.iter
      (fun block ->
	 let k = Int64.to_int(Int64.rem block bmaprowsizeL) in
	 let cok = Bitv.get crow k in
	 let rok = Bitv.get rrow k && Bitv.get orow_or k in
	 (** if it is in [rrow] but not in [orow_or] it is reserved by a
             different owner
	  *)
	 if not cok && not rok then
	   failwith
	     (sprintf 
		"Nn_state.blockmap.free: cannot free this block (id=%s b=%Ld)"
		identity block);
	 assert(cok <> rok);  (* not both! *)
	 if rok then (
	   (** Case: We allocated the block earlier in this transaction.
               We keep the block reserved. No need to modify [orow_and]
               as the block is still free in [committed_map].
	    *)
	   assert(not(Bitv.get orow_res k));
	   Bitv.set orow_or k false;
	   decr n_or;    (* number of 1 bits in orow_or *)
	 );
	 if cok then (
	   (** Case: The block is allocated in the committed view. No
               need to set it as reserved as it is allocated anyway.
               Only modify [orow_and] so it will be freed at commit time
	    *)
	   assert(Bitv.get orow_res k);
	   assert(not (Bitv.get rrow k));
	   Bitv.set orow_and k false;
	   incr n_and;   (* number of 0 bits in orow_and *)
	 );
	 (** If we set this block as to keep earlier, we undo this now. *)
	 if Bitv.get orow_keep k then (
	   decr n_keep;
	   Bitv.set orow_keep k false;
	 )
      )
      l
      (** [free_at] does not change which block is reserved, so
          [alloctab] remains unchanged
       *)

  method keep l ownobj =
    (** well, a dumb loop. FIXME *)
    if not active then raise Inactive;
    self # fix_alloctab();
    List.iter
      (fun block ->
	 let row = Int64.div block bmaprowsizeL in
	 if row >= size_rows_64 then
	   invalid_arg "Nn_state.blockmap.free";
	 self # keep_at (Int64.to_int row) [block] ownobj
      )
      l

  method private keep_at row l ownobj =
    (** Keeps the blocks in [l] which are all in [row] *)
    let crow = committed_map.(row) in
    let rrow = reserve_map.(row) in
    let krow = keep_map.(row) in
    let (orow_res, orow_keep, orow_or, orow_and, 
	 n_res, n_keep, n_or, n_and) =
      self # get_owner_tuple row ownobj in
    List.iter
      (fun block ->
	 let k = Int64.to_int(Int64.rem block bmaprowsizeL) in
	 let cok = Bitv.get crow k in
	 let rok = Bitv.get rrow k && Bitv.get orow_or k in
	 (** if it is in [rrow] but not in [orow_or] it is reserved by a
             different owner
	  *)
	 if not cok && not rok then
	   failwith
	     (sprintf 
		"Nn_state.blockmap.free: cannot keep this block (id=%s b=%Ld)"
		identity block);
	 assert(cok <> rok);  (* not both! *)
	 (** Case: We allocated the block earlier in this transaction.
             We keep the block reserved. It is not required to set the
	     keep bit - other transaction cannot see this block yet.
	     So: [ if rok then <do nothing> ]
	  *)
	 if cok then (
	   (** Case: The block is allocated in the committed view. We have
	       to set it as "to keep". It is an error if it was freed before.
	    *)
	   assert(Bitv.get orow_res k);
	   assert(not (Bitv.get rrow k));
	   if not(Bitv.get orow_and k) then
	     failwith
	       (sprintf 
		  "Nn_state.blockmap.free: cannot keep this block (id=%s b=%Ld)"
		  identity block);
	   if not(Bitv.get orow_keep k) then (
	     Bitv.set orow_keep k true;
	     incr n_keep;   (* number of 1 bits in orow_keep *)
	   );
	   Bitv.set krow k true
	 );
      )
      l
      (** [keep_at] does not change which block is reserved, so
          [alloctab] remains unchanged
       *)
    
  method get_changes ownobj =
    if not active then raise Inactive;
    self # fix_alloctab();
    let orow_ht_opt =
      try Some(Hashtbl.find owner ownobj)
      with Not_found -> None in
    match orow_ht_opt with
      | None ->
	  []
      | Some orow_ht ->
	  let changes = ref [] in
	  Hashtbl.iter
	    (fun row (orow_res, orow_keep, orow_or, orow_and, _, _, _, _) ->
	       let crow = committed_map.(row) in
	       let crow' =
		 Bitv.bw_and (Bitv.bw_or crow orow_or) orow_and in
	       let s =
		 Bitv.to_string crow' in
	       let block = Int64.mul (Int64.of_int row) bmaprowsizeL in
	       changes := (block, s) :: !changes
	    )
	    orow_ht;
	  !changes


  method commit ownobj =
    if not active then raise Inactive;
    self # fix_alloctab();
    let orow_ht_opt =
      try Some(Hashtbl.find owner ownobj)
      with Not_found -> None in
    match orow_ht_opt with
      | None ->
	  ()
      | Some orow_ht ->
	  self # remove_owner ownobj;
	  Hashtbl.iter
	    (fun row (orow_res, orow_keep, orow_or, orow_and, 
		      n_res, n_keep, n_or, n_and) ->
	       let rownlist =
		 try Hashtbl.find rowowner row with Not_found -> [] in
	       let crow = committed_map.(row) in
	       let rrow = reserve_map.(row) in
	       let (u,r,fk,al) = stats.(row) in
	       let n = Bitv.length crow in
	       (* new committed row: add the new bits from orow_or, clear
		  bits via orow_and
		*)
	       let crow' = Bitv.bw_and (Bitv.bw_or crow orow_or) orow_and in
	       (* new reserved row: clear bits from orow_res *)
	       let rrow' =
		 Bitv.bw_and rrow orow_res in
	       (* new keep row: this is the OR of all other orow_keep rows *)
	       let krow' =
		 List.fold_left
		   (fun acc rownobj ->
		      try Bitv.bw_or acc (self # orow_keep rownobj row)
		      with Not_found -> acc
		   )
		   (Bitv.create n false)
		   rownlist in
	       (* for computing fk' we need this: *)
	       let fkrow' = Bitv.bw_and krow' (Bitv.bw_not crow') in
	       (* update stats: *)
	       let fk' = count_ones fkrow' in
	       let u' = count_ones crow'  in
		 (* no better way - there may be several owners freeing the
		    same block
		  *)
	       let r' = r - !n_res in
	       let al' = n - u' - r' - fk' in
	       (* store *)
	       committed_map.(row) <- crow';
	       reserve_map.(row) <- rrow';
	       keep_map.(row) <- krow';
	       stats.(row) <- (u',r',fk',al');
	       n_used <- Int64.add n_used (Int64.of_int (u' - u));
	       n_reserved <- Int64.add n_reserved (Int64.of_int (r' - r));
	       n_free_kept <- Int64.add n_free_kept (Int64.of_int (fk' - fk));
	       alloctab.( al ) <- IntSet.remove row alloctab.( al );
	       alloctab.( al' ) <- IntSet.add row alloctab.( al' );
	    )
	    orow_ht

  method rollback ownobj =
    if not active then raise Inactive;
    self # fix_alloctab();
    let orow_ht_opt =
      try Some(Hashtbl.find owner ownobj)
      with Not_found -> None in
    match orow_ht_opt with
      | None ->
	  ()
      | Some orow_ht ->
	  self # remove_owner ownobj;
	  Hashtbl.iter
	    (fun row (orow_res, orow_keep, orow_or, orow_and, 
		      n_res, n_keep, n_or, n_and) ->
	       (** No need to modify [committed_map]. In [reserve_map],
                   undo all reservations. Also update [keep_map].
		*)
	       let rownlist =
		 try Hashtbl.find rowowner row with Not_found -> [] in
	       let crow = committed_map.(row) in
	       let rrow = reserve_map.(row) in
	       let (u,r,fk,al) = stats.(row) in
	       let n = Bitv.length crow in
	       (* new reserved row: clear bits from orow_res *)
	       let rrow' =
		 Bitv.bw_and rrow orow_res in
	       (* new keep row: this is the OR of all other orow_keep rows *)
	       let krow' =
		 List.fold_left
		   (fun acc rownobj ->
		      try Bitv.bw_or acc (self # orow_keep rownobj row)
		      with Not_found -> acc
		   )
		   (Bitv.create n false)
		   rownlist in
	       (* for computing fk' we need this: *)
	       let fkrow' = Bitv.bw_and krow' (Bitv.bw_not crow) in
	       (* update stats: *)
	       let fk' = count_ones fkrow' in
	       let u' = u in
	       let r' = r - !n_res in
	       let al' = n - u' - r' - fk' in
	       (* store *)
	       reserve_map.(row) <- rrow';
	       keep_map.(row) <- krow';
	       stats.(row) <- (u',r',fk',al');
	       n_reserved <- Int64.add n_reserved (Int64.of_int (r' - r));
	       n_free_kept <- Int64.add n_free_kept (Int64.of_int (fk' - fk));
	       alloctab.( al ) <- IntSet.remove row alloctab.( al );
	       alloctab.( al' ) <- IntSet.add row alloctab.( al' );
	    )
	    orow_ht

  method inactivate() =
    active <- false;
    committed_map <- [| |];
    reserve_map <- [| |];
    keep_map <- [| |];
    alloctab <- [| |];
    Hashtbl.clear owner;
    Hashtbl.clear rowowner

end


let blockmap_from_db_e ~id ~identity ~size conn esys =
  let bm = 
    new blockmap ~id ~identity ~size in
  new Uq_engines.seq_engine
    (Nn_db.blockalloc_list_e
       ~datastore:id
	conn
       esys)
    (fun rows ->
       List.iter
	 (fun (index,data) ->
	    bm#fill_blockmap index data
	 )
	 rows;
       new Uq_engines.epsilon_engine (`Done bm) esys
    )
