(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_notify.ml 239 2010-06-23 16:49:03Z gerd $ *)

module I64Map = Map.Make(Int64)
open Plasma_util.Operators

type entry = 
    { id : int;
      inode : int64;
      exp : int64;
      destination : string
    }

module ESet = Set.Make(struct 
			 type t = entry 
			 let compare x1 x2 = compare x1.id x2.id
		       end)
  
type notifications =
    { mutable by_inode : ESet.t I64Map.t;
      mutable by_exp : ESet.t I64Map.t;
      mutable next_id : int;
      mutable clients : (Unix.sockaddr, Rpc_proxy.ManagedClient.t) Hashtbl.t;
      mutable expiring : bool;
    }

module Nots_var =
  Netplex_cenv.Make_var_type(struct type t = notifications end)

let notsname = "current_notifications"

let get_nots() =
  Nots_var.get notsname

module Nots_proxy =
  Pfs_rpcapi_clnt.Make'Notifications(Rpc_proxy.ManagedClient)

let remove_entry nots e =
  ( try
      let el1 = I64Map.find e.inode nots.by_inode in
      let el2 = ESet.remove e el1 in
      if el2 = ESet.empty then
	nots.by_inode <- I64Map.remove e.inode nots.by_inode
      else
	nots.by_inode <- I64Map.add e.inode el2 nots.by_inode
    with Not_found -> ()
  );
  ( try
      let el1 = I64Map.find e.exp nots.by_exp in
      let el2 = ESet.remove e el1 in
      if el2 = ESet.empty then
	nots.by_exp <- I64Map.remove e.exp nots.by_exp
      else
	nots.by_exp <- I64Map.add e.exp el2 nots.by_exp
    with Not_found -> ()
  )


let add_entry nots e =
  let el1 =
    try I64Map.find e.inode nots.by_inode
    with Not_found -> ESet.empty in
  nots.by_inode <- I64Map.add e.inode (ESet.add e el1) nots.by_inode;
  let el2 =
    try I64Map.find e.exp nots.by_exp
    with Not_found -> ESet.empty in
  nots.by_exp <- I64Map.add e.exp (ESet.add e el2) nots.by_exp


let create_client conn =
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let mclient_config =
    Rpc_proxy.ManagedClient.create_mclient_config
      ~programs:[ Pfs_rpcapi_clnt.Notifications.V1._program ]
      ~msg_timeout:5.0
      ~msg_timeout_is_fatal:true
      ~idle_timeout:5.0
      ~initial_ping:true
      () in
  let mclient =
    Rpc_proxy.ManagedClient.create_mclient
      mclient_config
      conn
      esys in
  mclient


let get_client nots saddr =
  try
    Hashtbl.find nots.clients saddr
  with
    | Not_found ->
	let c = create_client (Plasma_util.connector_of_sockaddr saddr) in
	Hashtbl.add nots.clients saddr c;
	c


let parse_sockaddr s =
  if s <> "" && s.[0] = '/' then
    Unix.ADDR_UNIX s
  else
    Plasma_util.sockaddr_of_host_port (Plasma_util.parse_host_port s)


let send_notifications_e ml =
  (** We are only interested in certain db modifications, so look for that
      first
   *)
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let nots = get_nots() in
  let entries = ref ESet.empty in
  List.iter
    (function
       | `Inode_upd(inode,_)
       | `Inode_upd_time(inode,_,_)
       | `Inode_del inode ->
	   ( try
	       let el = I64Map.find inode nots.by_inode in
	       entries := ESet.union !entries el
	     with
	       | Not_found -> ()
	   )
       | _ -> ()
    )
    ml;
  (** Send out notifications: *)
  let engines =
    ESet.fold
      (fun e acc ->
	 let saddr = parse_sockaddr e.destination in
	 let c = get_client nots saddr in
	 let e =
	   Plasma_util.rpc_engine 
	     c Nots_proxy.V1.inode_update'async(e.inode, false)
	   >> (function
		 | `Done () -> `Done ()
		 | `Error err ->
		     Netlog.logf `Err
		       "Exception calling inode_update on %s: %s"
		       e.destination
		       (Netexn.to_string err);
		     `Done()
	       | `Aborted -> `Aborted
	      ) in
	 e :: acc
      )
      !entries
      [] in

  Plasma_util.msync_engine engines esys


let expire_e () =
  let now = Int64.of_float(Unix.time()) in
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let nots = get_nots() in

  if nots.expiring then (
    Netlog.logf `Warning
      "Nn_notify.expire_e: Last expiration cycle still active";
    eps_e (`Done()) esys
  )
  else (
    nots.expiring <- true;
 
    (** Extract the entries that expire by now: *)
    let entries = ref ESet.empty in
    ( try
	I64Map.iter
	  (fun t el ->
	     if t > now then raise Exit;
	     entries := ESet.union !entries el
	  )
	  nots.by_exp
      with
	| Exit -> ()
    );
    
    (** Really remove the entries: *)
    ESet.iter (remove_entry nots) !entries;

    (** Send out notifications: *)
    let engines =
      ESet.fold
	(fun e acc ->
	   let saddr = parse_sockaddr e.destination in
	   let c = get_client nots saddr in
	   let e =
	     Plasma_util.rpc_engine
	       c Nots_proxy.V1.inode_update'async(e.inode, true)
	     >> (function
		   | `Done () -> `Done ()
		   | `Error err ->
		       Netlog.logf `Err
			 "Exception calling inode_update on %s: %s"
			 e.destination
			 (Netexn.to_string err);
		       `Done()
		   | `Aborted -> `Aborted
		) in
	   e :: acc
	)
	!entries
	[] in
    
    Plasma_util.msync_engine engines esys
    ++ (fun () ->
	  nots.expiring <- false;
	  eps_e (`Done()) esys
       )
  )

let post_start_hook() =
  Nots_var.set
    notsname
    { by_inode = I64Map.empty;
      by_exp = I64Map.empty;
      next_id = 0;
      clients = Hashtbl.create 5;
      expiring = false;
    }


let proc_on_inode_update (inode,exp,destination) =
  let nots = get_nots() in
  let now = Int64.of_float(Unix.time()) in
  let exp_ok = (exp > now) in
  let dest_ok =
    ( try ignore(parse_sockaddr destination); true with _ -> false ) in
  let ok = exp_ok && dest_ok in

  if ok then (
    let e =
      { id = nots.next_id;
	inode = inode;
	exp = exp;
	destination = destination;
      } in
    nots.next_id <- nots.next_id + 1;
    add_entry nots e;
  );

  ok


let setup srv =
  Pfs_rpcapi_srv.Request_notifications.V1.bind
    ~proc_null:(fun () -> ())
    ~proc_on_inode_update
    srv
